using System;
using System.Collections.Generic;
using System.Text;

namespace Implicit2D
{
    public struct Box2f
    {
        public Vector2f Min;
        public Vector2f Max;

        public Box2f(bool bIgnore) {
            Min = new Vector2f(Single.MaxValue, Single.MaxValue);
            Max = new Vector2f(Single.MinValue, Single.MinValue);
        }

        public Box2f(float xmin, float ymin, float xmax, float ymax) {
            Min = new Vector2f(xmin, ymin);
            Max = new Vector2f(xmax, ymax);
        }

        public Box2f(float fWidth, float fHeight) {
            Min = new Vector2f(0, 0);
            Max = new Vector2f(fWidth, fHeight);
        }

        public Box2f(Vector2f vMin, Vector2f vMax) {
            Min.X = Math.Min(vMin.X, vMax.X);
            Max.X = Math.Max(vMin.X, vMax.X);
            Min.Y = Math.Min(vMin.Y, vMax.Y);
            Max.Y = Math.Max(vMin.Y, vMax.Y);
        }

        public Box2f(Box2f o) {
            Min = new Vector2f(o.Min);
            Max = new Vector2f(o.Max);
        }

        public float Width {
            get { return Max.X - Min.X; }
        }
        public float Height {
            get { return Max.Y - Min.Y; }
        }

        public float Area {
            get { return Width * Height; }
        }
        public float DiagonalLen {
            get { return (float)Math.Sqrt((Max.X - Min.X) * (Max.X - Min.X) + (Max.Y - Min.Y) * (Max.Y - Min.Y)); }
        }
        public float MaxDim {
            get { return Math.Max(Width, Height); }
        }

        public Vector2f Center {
            get { return new Vector2f(0.5f * (Min.X + Max.X), 0.5f * (Min.Y + Max.Y)); }
        }


        //! value is subtracted from min and added to max
        public void Expand(float fRadius) {
            Min.X -= fRadius; Min.Y -= fRadius;
            Max.X += fRadius; Max.Y += fRadius;
        }
        //! value is added to min and subtracted from max
        public void Contract(float fRadius) {
            Min.X += fRadius; Min.Y += fRadius;
            Max.X -= fRadius; Max.Y -= fRadius;
        }
        // values are all added
        public void Pad(float fPadLeft, float fPadRight, float fPadBottom, float fPadTop) {
            Min.X += fPadLeft; Min.Y += fPadBottom;
            Max.X += fPadRight; Max.Y += fPadTop;
        }

        public enum ScaleMode {
            ScaleRight,
            ScaleLeft,
            ScaleUp,
            ScaleDown,
            ScaleCenter
        }
        public void SetWidth( float fNewWidth, ScaleMode eScaleMode ) {
            switch (eScaleMode) {
                case ScaleMode.ScaleLeft:
                    Min.X = Max.X - fNewWidth;
                    break;
                case ScaleMode.ScaleRight:
                    Max.X = Min.X + fNewWidth;
                    break;
                case ScaleMode.ScaleCenter:
                    Vector2f vCenter = Center;
                    Min.X = vCenter.X - 0.5f * fNewWidth;
                    Max.X = vCenter.X + 0.5f * fNewWidth;
                    break;
                default:
                    throw new Exception("Invalid scale mode...");
            }
        }
        public void SetHeight(float fNewHeight, ScaleMode eScaleMode) {
            switch (eScaleMode) {
                case ScaleMode.ScaleDown:
                    Min.Y = Max.Y - fNewHeight;
                    break;
                case ScaleMode.ScaleUp:
                    Max.Y = Min.Y + fNewHeight;
                    break;
                case ScaleMode.ScaleCenter:
                    Vector2f vCenter = Center;
                    Min.Y = vCenter.Y - 0.5f * fNewHeight;
                    Max.Y = vCenter.Y + 0.5f * fNewHeight;
                    break;
                default:
                    throw new Exception("Invalid scale mode...");
            }
        }

        public void Contain(Vector2f v) {
            Min.X = Math.Min(Min.X, v.X);
            Min.Y = Math.Min(Min.Y, v.Y);
            Max.X = Math.Max(Max.X, v.X);
            Max.Y = Math.Max(Max.Y, v.Y);
        }

        public void Contain(Box2f box) {
            Contain(box.Min);
            Contain(box.Max);
        }

        public Box2f Intersect(Box2f box) {
            return new Box2f(
            Math.Max(Min.X, box.Min.X), Math.Max(Min.Y, box.Min.Y),
            Math.Min(Max.X, box.Max.X), Math.Min(Max.Y, box.Max.Y));
        }



        public bool Contains(Vector2f v) {
            return (Min.X < v.X) && (Min.Y < v.Y) && (Max.X > v.X) && (Max.Y > v.Y);
        }
        public bool Intersects(Box2f box) {
            return !((box.Max.X < Min.X) || (box.Min.X > Max.X) || (box.Max.Y < Min.Y) || (box.Min.Y > Max.Y));
        }


        //! relative translation
        public void Translate(Vector2f vTranslate) {
            Min.Add(vTranslate);
            Max.Add(vTranslate);
        }

        public void MoveMin(Vector2f vNewMin) {
            Max.X = vNewMin.X + (Max.X - Min.X);
            Max.Y = vNewMin.Y + (Max.Y - Min.Y);
            Min.Set(vNewMin);
        }
        public void MoveMin(float fNewX, float fNewY) {
            Max.X = fNewX + (Max.X - Min.X);
            Max.Y = fNewY + (Max.Y - Min.Y);
            Min.Set(fNewX, fNewY);
        }

    }
}
